#!/usr/bin/env python
# -*- coding: utf-8 -*-
try:
    from django.db.backends import util as utils
except ImportError:
    from django.db.backends import utils
from django.db.backends.postgresql_psycopg2 import base

from production_request.utils import calc_sql_metrics


class CursorWrapper(utils.CursorWrapper):
    """
    Враппер курсора БД для подсчета времени выполнения запроса
    """
    def execute(self, sql, params=None):
        with calc_sql_metrics(self, sql):
            return self.cursor.execute(sql, params)

    def executemany(self, sql, param_list):
        with calc_sql_metrics(self, sql):
            return self.cursor.executemany(sql, param_list)


class DatabaseWrapper(base.DatabaseWrapper):
    """
    Враппер БД с кастомным курсором
    """
    def cursor(self):
        if hasattr(self, 'validate_thread_sharing'):
            self.validate_thread_sharing()
        cursor = CursorWrapper(self._cursor(), self)
        return cursor

    def _commit(self):
        with calc_sql_metrics(self.cursor(), 'COMMIT'):
            return super(DatabaseWrapper, self)._commit()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
import uuid
from datetime import datetime
from psutil import Process

from django.conf import settings
from django.db import connections

from m3_legacy.middleware import get_thread_data

from production_request.enums import SqlOperationEnum
from production_request.utils import (
    ms_from_timedelta, get_client_ip, register_logger)


PRODUCTION_REQUEST_LOGGER = register_logger(
    'production_request',
    formatter=logging.Formatter("%(message)s"))


class ProductionRequestLoggingMiddleware(object):
    """
    Middleware для логирования запросов на production.
    Использует данные из production_request.database.CursorWrapper
    """
    # Атрибуты, не добавляемые в хэдеры
    exclude_attr_headers = ('uuid',)

    def process_request(self, request):
        self._start_log_request(request)

    def process_response(self, request, response):
        log_row = self._log_request(request)
        self._add_response_headers(request, response, log_row)

        return response

    def process_exception(self, request, exception):
        self._log_request(request)

    def _start_log_request(self, request):
        """
        Начинает процесс логирования запроса
        """
        if settings.PRODUCTION_REQUEST:
            request_uuid = str(uuid.uuid4())[-12:]

            thread_data = get_thread_data()
            thread_data.production_request_uuid = request_uuid

            request.started_at = datetime.now()
            request.production_request_uuid = request_uuid

            if settings.PRODUCTION_REQUEST_LOG_MEMORY:
                memory = self._get_memory_info()

                request.production_request_uss = memory.uss
                request.production_request_pss = memory.pss
                request.production_request_swap = memory.swap
                request.production_request_rss = memory.rss

    def _get_memory_info(self):
        """Возвращает данные о памяти процесса"""
        return Process(os.getpid()).memory_full_info()

    def _get_memory_diff(self, m_begin, m_end):
        """Возвращает разницу размера памяти в МБ"""
        return (m_end - m_begin) / 1024 / 1024 / 8

    def _log_request(self, request):
        """Логирует параметры запроса и результаты обработки"""
        row = None
        if settings.PRODUCTION_REQUEST:
            try:
                _uuid = request.META.get(
                    'HTTP_PR_UUID',
                    ''
                )
                process_request_time = ms_from_timedelta(
                    datetime.now() - request.started_at
                )
                process_sql_time = connections['default'].sql_time.pop(
                    request.production_request_uuid, float())

                sql_count = connections['default'].sql_count.pop(
                    request.production_request_uuid, 0)

                row = {
                    'uuid': _uuid,
                    'started': str(
                        request.started_at
                    ),
                    'path': request.path,
                    'user': str(request.user.id or ''),
                    'total': '{:.4f}'.format(process_request_time),
                    'sql_total': '{:.4f}'.format(process_sql_time),
                    'sql_count': str(sql_count),
                    'client_ip': get_client_ip(
                        request
                    ),
                    'pid': str(os.getpid())
                }

                # количестов sql-запросов в разрезе типов
                for operation in SqlOperationEnum.values:
                    operation_attr = 'sql_{}_count'.format(operation)
                    counter = getattr(
                        connections['default'],
                        operation_attr,
                        {}
                    )
                    row[operation_attr] = counter.pop(
                        request.production_request_uuid, 0
                    )

                if settings.PRODUCTION_REQUEST_LOG_MEMORY:
                    memory = self._get_memory_info()

                    row['uss'] = '{:.2f}'.format(
                        self._get_memory_diff(
                            request.production_request_uss,
                            memory.uss
                        )
                    )
                    row['pss'] = '{:.2f}'.format(
                        self._get_memory_diff(
                            request.production_request_pss,
                            memory.pss
                        )
                    )
                    row['swap'] = '{:.2f}'.format(
                        self._get_memory_diff(
                            request.production_request_swap,
                            memory.swap
                        )
                    )
                    row['rss'] = '{:.2f}'.format(
                        self._get_memory_diff(
                            request.production_request_rss,
                            memory.rss
                        )
                    )

                if settings.PRODUCTION_REQUEST_LOG_SERVER:
                    PRODUCTION_REQUEST_LOGGER.info(
                        row
                    )
            except Exception as err:
                if settings.PRODUCTION_REQUEST_LOG_SERVER:
                    PRODUCTION_REQUEST_LOGGER.error(
                        str(
                            err
                        )
                    )

            return row

    def _create_header(self, attr):
        return 'PR-S-{}'.format(attr.upper().replace("_", "-"))

    def _add_response_headers(self, request, response, log_row):
        if settings.PRODUCTION_REQUEST_LOG_CLIENT:
            try:
                _uuid = request.META.get(
                    'HTTP_PR_UUID',
                )
                _started = request.META.get(
                    'HTTP_PR_STARTED',
                )

                if _uuid and _started:
                    response['PR-UUID'] = _uuid
                    response['PR-STARTED'] = _started

                if log_row:
                    for key, value in log_row.items():
                        if key not in self.exclude_attr_headers:
                            response[self._create_header(key)] = value

            except Exception as err:
                if settings.PRODUCTION_REQUEST_LOG_SERVER:
                    PRODUCTION_REQUEST_LOGGER.error(str(err))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from logging.handlers import TimedRotatingFileHandler
import os
from collections import defaultdict
from contextlib import contextmanager
from datetime import datetime

from django.conf import settings

from m3_legacy.middleware import get_thread_data

from production_request.enums import SqlOperationEnum


def ms_from_timedelta(td):
    """
    Получает дельту, возвращает миллисекунды
    """
    return (td.seconds * 1000) + (td.microseconds / 1000.0)


@contextmanager
def calc_sql_metrics(cursor, sql):
    """
    Контекстный менеджер для подсчета времени выполнения SQL-запроса
    :param cursor: курсок к БД
    """
    start = datetime.now()

    yield

    request_uuid = getattr(
        get_thread_data(), 'production_request_uuid', None)
    if request_uuid:
        # подсчет времени выполнения sql-запроса
        if not hasattr(cursor.db, 'sql_time'):
            cursor.db.sql_time = defaultdict(float)

        duration = ms_from_timedelta(datetime.now() - start)
        cursor.db.sql_time[request_uuid] += duration

        # подсчет количества sql-запросов
        if not hasattr(cursor.db, 'sql_count'):
            cursor.db.sql_count = defaultdict(int)

        cursor.db.sql_count[request_uuid] += 1

        # подсчет количества sql-запросов определенного типа
        sql_operation = SqlOperationEnum.get_key_by_sql(sql)
        if sql_operation:
            sql_operation_attr = 'sql_{}_count'.format(sql_operation)

            if not hasattr(cursor.db, sql_operation_attr):
                setattr(cursor.db, sql_operation_attr, defaultdict(int))

            getattr(cursor.db, sql_operation_attr)[request_uuid] += 1
        else:
            print(
                'PRODUCTION-REQUEST: Unknown SQL-operation. '
                'Will be added to total count only.'
            )


def get_client_ip(request):
    """
    Возвращает IP-адрес клиента, от которого пришел запрос
    """
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR', '')

    return ip


def register_logger(
        lname, log_file=None, log_level=logging.INFO, formatter=None):
    """
    Регистрирует логгер
    """
    log_file = log_file or u'{0}.log'.format(lname)
    log_dir = settings.LOG_PATH
    log_path = os.path.join(log_dir, log_file)

    if formatter is None:
        formatter = logging.Formatter(
            "\n%(pathname)s:%(lineno)d\n[%(asctime)s] %(levelname)s: "
            "%(message)s")
    formatter.datefmt = '%Y-%m-%d %H:%M:%S'

    l = logging.getLogger(lname)
    l.setLevel(log_level)

    # Защита от reload()
    if not l.handlers:
        handler = TimedRotatingFileHandler(
            log_path, when='D', encoding='utf-8')
        handler.setFormatter(formatter)
        l.addHandler(handler)

    return l




#!/usr/bin/env python
# -*- coding: utf-8 -*-


class SqlOperationEnum(object):
    u"""Типы операций в sql"""
    UNKNOWN = ''
    SELECT = 's'
    INSERT = 'i'
    UPDATE = 'u'
    DELETE = 'd'
    COMMIT = 'c'
    SAVEPOINT = 'sp'

    values = {
        SELECT: 'SELECT',
        INSERT: 'INSERT',
        UPDATE: 'UPDATE',
        DELETE: 'DELETE',
        COMMIT: 'COMMIT',
        SAVEPOINT: 'SAVEPOINT',
    }

    @classmethod
    def get_key_by_sql(cls, sql):
        u"""
        Возвращает ключ операции по sql-запросу
        :param sql: sql-запрос
        :rtype: str
        """
        for key, value in cls.values.items():
            if value in sql:
                result = key
                break
        else:
            result = cls.UNKNOWN

        return result


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from django.conf import settings

try:
    from m3 import OperationResult
    from m3.actions import ActionPack, ACD, Action
except ImportError:
    from m3.ui.actions import ActionPack, ACD, Action, OperationResult

from production_request.utils import register_logger


PRODUCTION_REQUEST_CLIENT_LOGGER = register_logger(
    'production_request_client',
    formatter=logging.Formatter("%(message)s"))


class ProductionRequestPack(ActionPack):
    """
    Пак для работы с логами production_request
    """

    url = '/production-request'

    def __init__(self):
        super(ProductionRequestPack, self).__init__()

        self.action_save_client_log = SaveLogAction()

        self.actions.extend([
            self.action_save_client_log,
        ])


class SaveLogAction(Action):
    """
    Сохранение лога
    """
    url = '/save-client-log'

    def context_declaration(self):
        return [
            ACD('logs', type=object, required=True, default=[]),
        ]

    def run(self, request, context):
        if settings.PRODUCTION_REQUEST_LOG_CLIENT:
            for log_str in context.logs:
                PRODUCTION_REQUEST_CLIENT_LOGGER.info(log_str)

        return OperationResult()
